/*
  mkfs.c
  ------
  $Id: mkfs.c,v 1.23 2003/11/03 19:29:16 stewart Exp $
  (C)2003 Stewart Smith
  Distributed under the GNU Public License

  This is the 'mkfs' utility for FCFS - the new Walnut object store.
  The name mkfs is kept purely 'cause it says what it does.

  Some data structures have been constructed out of those
  present in the Linux Kernel (v2.5.69). They are copyright
  of their respective owners.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>


#include "testkit/block_dev.h"
#include "testkit/types.h"
#include "testkit/bitops.h"
#include "disk.h"

#include "super_block.h"
#include "onode.h"
#include "onode_index.h"
#include "space_bitmap.h"

#define EXPERIMENTAL

u64 atou64(const char *nptr)
{
  u64 out,i;
  for(out=0,i=0;nptr[i]!='\0';i++)
    {
      out*=10;
      out+=nptr[i]-'0';
    }

  return out;
}

int make_sb(void *sb_v, u32 bsize,u64 bcount, char* name)
{
  struct fcfs_sb *sb = (struct fcfs_sb*) sb_v;

  sb->magic1 = FCFS_SB_MAGIC1;
  sb->version = FCFS_SB_VERSION1;
  sb->bits = 64;
  sb->sblength = sizeof(struct fcfs_sb);
  sb->flags = 0;
  set_bit(FCFS_FLAG_JournalMeta,&(sb->flags));
#ifdef EXPERIMENTAL
  set_bit(FCFS_FLAG_Experimental,&(sb->flags));
#endif

  sb->bsize = bsize;
  sb->blocksnr = bcount;
  strcpy(sb->name,name);
  sb->magic2 = FCFS_SB_MAGIC2;

  sb->time_created = time(NULL);
  sb->time_clean = time(NULL);
  sb->time_dirtymount = 0x0ULL;
  
  sb->num_mounts = 0x0ULL;
  sb->num_dirtymounts = 0x0ULL;

  /* We simply steal the number of allocation groups from XFS. :)
     They use 8, so we'll steal that.
   */
  sb->allocation_groupsnr = 8;
  sb->ag_blocksnr = bcount / sb->allocation_groupsnr;

  fprintf(stderr,"ALLOCATION GROUPS: 0x%x, each 0x%x\n",
	  sb->allocation_groupsnr,
	  sb->ag_blocksnr);

  sb->onindex_num_onodes = 0;	/* Num onodes in index */
  sb->onindex_used = 0;		/* Num of used inodes in index */
  sb->onindex_free = 0;		/* Num of free onodes in index */
  sb->onindex_next_id = 0;	/* Next ID to use */

  /* FIXME: Potentially do real nasty things that can't be described
   * in polite circles when our structs change size... */
  sb->onindex_node_size =  ((sb->bsize - 4*sizeof(u64)) / sizeof(u64))/2;
  /* In number of keys */

  return 1;			/* success, although we don't check it*/
}

int write_superblocks(struct fcfs_disk* disk)
{
  struct fcfs_disk_block *block;
  u32 i;

  /* Write start of volume master SB */
  block = disk_getblock(disk,0);

  disk->sb->flags = disk->sb->flags | FCFS_SBloc_start_volume;  /* start of volume*/

  disk_writeblock(block);
  disk_freeblock(block);
  fprintf(stderr,"SuperBlock at Block 0\n");
  
  /* Write end of volume Backup SB */
  block = disk_newblock(disk,disk->blocksnr-1ULL);


  //  disk->sb->flags = disk->sb->flags & ~0x03ULL;
  disk->sb->flags = disk->sb->flags ^ FCFS_SBloc_start_volume;  /* start of volume*/
  disk->sb->flags = disk->sb->flags | FCFS_SBloc_end_volume;  /* end of volume*/

  memcpy(block->data,disk->sb,disk->bsize);

  disk_writeblock(block);
  disk_freeblock(block);
  fprintf(stderr,"BACKUP SuperBlock at Block 0x%llx\n",disk->blocksnr-1ULL);

  disk->sb->flags = disk->sb->flags ^ FCFS_SBloc_end_volume;  /* end of volume*/

  for(i=1;i<disk->sb->allocation_groupsnr;i++)
    {
      block = disk_newblock(disk,i*disk->sb->ag_blocksnr);

      disk->sb->flags = disk->sb->flags | FCFS_SBloc_start_ag;  /* start of ag*/
      memcpy(block->data,disk->sb,disk->bsize);
      disk_writeblock(block);
      disk_freeblock(block);
      fprintf(stderr,"BACKUP SuperBlock at Block 0x%llx (offset 0x%llx)\n",i*disk->sb->ag_blocksnr,(i*disk->sb->ag_blocksnr)*disk->bsize);
    }
  return 0;
}

int write_block_bitmaps(struct fcfs_disk* disk)
{
  struct fcfs_disk_block *block;
  struct fcfs_block_run br;
  int i,j;

  /* Create initial empty Block Bitmap */
  for(i=0;i<(disk->sb->allocation_groupsnr);i++)
    {
      /* For each ag, blank one of the right size */
      for(j=0;j<space_bitmap_size(disk->sb,i);j++)
	{
	  block = disk_newblock(disk,i*disk->sb->ag_blocksnr+2);
	  memset(block->data,0,disk->bsize);
	  fprintf(stderr,"BITMAP: Zeroing 0x%x (offset %x) length %u\n",
		  i*disk->sb->ag_blocksnr+2,
		  (i*disk->sb->ag_blocksnr+2)*disk->bsize,
		  space_bitmap_size(disk->sb,i));
	  disk_writeblock(block);
	  disk_freeblock(block);
	}
    }

  for(i=0;i<(disk->sb->allocation_groupsnr);i++)
    {
      /* Setup our Block Run */
      br.allocation_group = i;
      br.start = 1;
      br.len = space_bitmap_size(disk->sb,i);

      space_bitmap_allocate_block(disk,i,0);

      for(j=0;j<space_bitmap_size(disk->sb,i);j++)
	space_bitmap_allocate_block(disk,i,j+1);

    }

  return 0;
}

int main(int argc,char* argv[])
{
  struct block_device bdev;
  struct fcfs_disk *disk;
  struct fcfs_disk_block *block;
  struct fcfs_sb *sb;
  struct fcfs_onode_index *index;
  u64 bsize;
  u64 bcount;
  char* buffer;
  int i;

  if(argc<5)
    {
      fprintf(stderr,"FCFS Make Object Store Utility (mkfs)\n");
      fprintf(stderr,"-------------------------------------\n");
      fprintf(stderr,"$Id: mkfs.c,v 1.23 2003/11/03 19:29:16 stewart Exp $\n\n");
      fprintf(stderr,"Written by: Stewart Smith (stewart@flamingspork.com)\n\n");
      fprintf(stderr,"Usage:\n\t./mkfs device blocksize blockcount name [initial objects]\n\n");
      exit(0);
    }

  bsize =  atou64(argv[2]);
  bcount = atou64(argv[3]);  

  fprintf(stderr,"Going to create volume '%s' with 0x%llx blocks at 0x%llx bytes each\n\n",argv[4],bcount,bsize);

  block_dev_init();
  block_dev_new(&bdev,argv[1],bsize,bcount);
  disk = disk_new(&bdev);

  if((buffer = (char*)malloc(sizeof(char*)*bsize))==0)
    {
      fprintf(stderr,"Unable to allocate buffer\n");
      abort();
    }

  /* Make a new Super Block */
  block = disk_newblock(disk,0);
  sb = (struct fcfs_sb*)block->data;

  if(!make_sb(sb,disk->bsize,disk->blocksnr,argv[4]))
    {
      fprintf(stderr,"Make_sb failed\n");
      abort();
    }

  disk->sb = sb;

  write_superblocks(disk);

  write_block_bitmaps(disk);

  index = onode_index_new(disk);
  onode_index_new_root(index);

  for(i=5;i<argc;i++)
  {
    struct fcfs_onode1 *node;
    struct fcfs_block_run *onode_br;
    int forknr;
    char data[1024]; int j;
    FILE* a;

    node = onode1_new(disk);
    onode_br = onode_index_insert(index,node);
    //    onode1_grow(disk,onode_br,10);
    //    onode1_grow(disk,onode_br,20);
    onode1_fork_new(disk,onode_br,0x42,strlen(argv[i]),argv[i],1);
    forknr = onode1_fork_new(disk,onode_br,0x69,0,NULL,0);
    fprintf(stderr,"+++++++GOING TO GO AND WRITE %s++++++++\n",argv[5]);
    a = fopen(argv[i],"r");
    j=0;
    while(!feof(a))
      {
	fgets(data,1000,a);
	fprintf(stderr,"*******WRITING %d bytes*******\n",strlen(data));
	onode1_fork_write(disk,onode_br,forknr,j,(u64)strlen(data),data);
	j+=strlen(data);
      }
    fclose(a);
  }  

  disk_freeblock(block);

  /* Clean up and exit */
  block_dev_close(&bdev);
  
  return 0;
}
