/* 
 super_block.h
 -------------
 $Id: super_block.h,v 1.13 2003/10/20 07:18:11 stewart Exp $
 
 Part of the FCFS object store.

 This file describes the super block and related data structures.

 (C)2003 Stewart Smith
 Distributed under the GNU Public License
*/
#ifndef __FCFS_SUPER_BLOCK_H__
#define __FCFS_SUPER_BLOCK_H__

#include "disk.h"

/* fcfs_block_run
   --------------
   We have a max of ~4billion alloction groups,
   within these, we can have up to 4billion blocks.
   and we can address these with *one* block_run structure.
   This means we don't have a lot of block address lookup on large files
   or contiguous files. Should lead to fast IO...
 */
struct fcfs_block_run {
  u32 allocation_group;		/* Allocation group */
  u32 start;			/* Start Block */
  u32 len;			/* Length (blocks) */
};

#define FCFS_SB_MAGIC1   0x46436673  /* FCFS */
#define FCFS_SB_VERSION1 0x00010001  /* V1 SB, V1 FS */
#define FCFS_SB_BITS     0x00000040  /* We're 64bit - but could be bigger...*/

#define FCFS_NAME_LENGTH 128	     /* Human Readable Name */
#define FCFS_SB_MAGIC2   0x3f8ec2a1  /* There is no meaning, just weird num */

enum fcfs_sb_flags {
  FCFS_FLAG_SBLocation1,	/* Three flags make number */
  FCFS_FLAG_SBLocation2,	/* see: enum fcfs_sb_location */
  FCFS_FLAG_SBLocation3,
  FCFS_FLAG_Dirty,		/* Was uncleanly unmounted*/
  FCFS_FLAG_Experimental,	/* Has been written by experimental code */
  FCFS_FLAG_JournalMeta,	/* Journal meta-data */
  FCFS_FLAG_JournalData,	/* Journal all data writes*/
  FCFS_FLAG_Versioned,		/* Is versioned system */
};

enum fcfs_sb_location {
  FCFS_SBloc_start_volume,	/* Start of volume */
  FCFS_SBloc_start_ag,		/* Start of allocation group */
  FCFS_SBloc_end_volume		/* Final block of volume */
};

struct fcfs_sb {
  u32 magic1;    // first magic number. 
  u32 version;   // version of FS
  u32 bits;      // number of bits to use as a 'base' bits. (64 default)
  u32 sblength;   // length of sb (bytes)

  u64 flags;

  u32 bsize;     // block size (bytes)
  u64 blocksnr;  // number of blocks


  char name[FCFS_NAME_LENGTH];  // human readable name of volume
  u32 magic2;    // magic2

  u32 allocation_groupsnr;	/* Number of alloction groups on volume */
  u32 ag_blocksnr;		/* Blocks per allocation group */

  u64 num_mounts;		/* Number of successful read/write mounts */
  u64 num_dirtymounts;		/* Num dirty mounts */
  u64 time_created;		/* Time volume created*/
  u64 time_clean;		/* Time volume last unmounted cleanly*/
  u64 time_dirtymount;		/* Time volume last mounted dirty */

  u64 onode_index_blocknr;	/* Where we find our primary index */

  u64 onindex_num_onodes;	/* Num onodes in index */
  u64 onindex_used;		/* Num of used inodes in index */
  u64 onindex_free;		/* Num of free onodes in index */
  u64 onindex_next_id;		/* Next ID to use */
  u32 onindex_node_size;	/* In number of keys */

  union space_tracking {
    
    char padding[50];
  }space;
};

#define FCFS_ONODE_MAGIC1  0x4f4e6f4465563101ULL	/* ONoDeV1 0x01 */

enum fcfs_onode_flags {
  FCFS_OFLAG_NoVersion,		/* Don't version track this onode */
  FCFS_OFLAG_ForkLeaf		/* We have a fork leaf, not a node */
};

#define FSFS_ONODE1_SPACE_LEAF_MAXNR 10

struct fcfs_onode1_space_leaf {
  char nr;
  struct fcfs_block_run br[FSFS_ONODE1_SPACE_LEAF_MAXNR];
};

struct fcfs_onode1_space_node {
  u64 offset[7];
  u64 block[7];
};

enum fcfs_fork_flags {
  FCFS_FORK_InForkData,		/* We have data, not space nodes */
  FCFS_FORK_SpaceNode		/* Space is node, not leaf */
};

#define FCFS_FORK_SMALL_DATA_SIZE 112

struct fcfs_fork {
  u64 fork_type;
  u64 fork_flags;
  u64 content_length;
  union data {
    union space {
      struct fcfs_onode1_space_leaf leaf;
      struct fcfs_onode1_space_node node;
    } space;
    
    char small_data[FCFS_FORK_SMALL_DATA_SIZE];	/* I wish this was neater, as part of the leaf */
  } data ;
};  

struct fcfs_fork_leaf {
  char nr;
  struct fcfs_fork fork[10];
};

struct fcfs_fork_node {
  u64 offset[7];
  u64 block[7];
};

struct fcfs_onode1 {
  u64 magic1;			/* Identify as O-Node */
  u64 onode_num;		/* FS Specific Unique ID */
  u64 onode_revision;		/* Revision of onode */
  u64 flags;			/* fcfs_onode_flags */
  u64 use_count;		/* Reference Counter (for indicies) */
  u32 onode_size;		/* Length of o-node structure. */
  union forks {
    struct fcfs_fork_leaf leaf;
    struct fcfs_fork_node node;
  } forks;
  char small_space[1];		/* used for in-onode data (or metadata) */
};


/* Functions for manipulating the Super Block */
int fcfs_write_sb(struct fcfs_disk *disk);
int fcfs_sb_mark_dirty(struct fcfs_disk *disk);
int fcfs_sb_mark_clean(struct fcfs_disk *disk);

#endif

